import sys, os

sys.path.append(os.getcwd())  # add current working directory to the path

from epimodel import EpidemiologicalParameters, run_model, preprocess_data
from epimodel.preprocessing import PreprocessedData
from epimodel.script_utils import *

import argparse
import json
import numpyro
import numpy as np
from datetime import datetime

argparser = argparse.ArgumentParser()

add_argparse_arguments(argparser)
argparser.add_argument(
    "--seed",
    dest="seed",
    type=int,
    help="seed for bootstrapping",
)

args = argparser.parse_args()

numpyro.set_host_device_count(args.num_chains)
if __name__ == "__main__":
    print(f"Running Sensitivity Analysis {__file__} with config:")
    config = load_model_config(args.model_config)
    pprint_mb_dict(config)

    print("Loading Data")
    data = preprocess_data(get_data_path())
    data.featurize(**config["featurize_kwargs"])
    data.mask_new_variant(
        new_variant_fraction_fname=get_new_variant_path(),
    )
    data.mask_from_date("2021-01-09")

    np.random.seed(args.seed)
    region_indices = np.random.choice(np.arange(data.nRs), size=data.nRs, replace=True)

    print(f"Boostrapped: {np.unique(region_indices).size} unique regions")

    Rs = np.array(data.Rs)[region_indices].tolist()
    new_cases = data.new_cases[region_indices, :]
    new_deaths = data.new_deaths[region_indices, :]
    Cs = np.array(data.Cs)[region_indices].tolist()
    active_cms = data.active_cms[region_indices, :, :]
    unique_Cs = sorted(list(set(Cs)))
    nCs = len(unique_Cs)

    C_indices = []
    for uc in unique_Cs:
        a_indices = np.nonzero([uc == c for c in Cs])[0]
        C_indices.append(a_indices)

    bootstrapped_data = PreprocessedData(
        Rs,
        data.Ds,
        data.CMs,
        new_cases,
        new_deaths,
        active_cms,
        Cs,
        unique_Cs,
        C_indices,
    )
    print("Loading EpiParam")
    ep = EpidemiologicalParameters()

    model_func = get_model_func_from_str(args.model_type)
    ta = get_target_accept_from_model_str(args.model_type)
    td = get_tree_depth_from_model_str(args.model_type)

    base_outpath = generate_base_output_dir(
        args.model_type, args.model_config, args.exp_tag
    )
    ts_str = datetime.now().strftime("%Y-%m-%d;%H:%M:%S")
    summary_output = os.path.join(base_outpath, f"{ts_str}_summary.json")
    full_output = os.path.join(base_outpath, f"{ts_str}_full.netcdf")

    model_build_dict = config["model_kwargs"]

    posterior_samples, _, info_dict, _ = run_model(
        model_func,
        bootstrapped_data,
        ep,
        num_samples=args.num_samples,
        num_chains=args.num_chains,
        num_warmup=args.num_warmup,
        target_accept=ta,
        max_tree_depth=td,
        model_kwargs=model_build_dict,
        save_results=True,
        output_fname=full_output,
        chain_method="parallel",
    )

    info_dict["model_config_name"] = args.model_config
    info_dict["model_kwargs"] = config["model_kwargs"]
    info_dict["featurize_kwargs"] = config["featurize_kwargs"]
    info_dict["start_dt"] = ts_str
    info_dict["exp_tag"] = args.exp_tag
    info_dict["exp_config"] = {
        "region_indices": region_indices.tolist(),
        "seed": args.seed,
    }
    info_dict["cm_names"] = data.CMs
    info_dict["data_path"] = get_data_path()

    # also need to add sensitivity analysis experiment options to the summary dict!
    summary = load_keys_from_samples(
        get_summary_save_keys(), posterior_samples, info_dict
    )
    with open(summary_output, "w") as f:
        json.dump(info_dict, f, ensure_ascii=False, indent=4)
